package checkers.jcip;

import java.util.List;

import javax.lang.model.element.AnnotationMirror;
import javax.lang.model.element.AnnotationValue;

import checkers.basetype.BaseTypeChecker;
import checkers.jcip.quals.GuardedBy;
import checkers.quals.TypeQualifiers;
import checkers.types.QualifierHierarchy;
import checkers.util.AnnotationUtils;
import checkers.util.GraphQualifierHierarchy;

@TypeQualifiers({ GuardedBy.class })
public class JCIPChecker extends BaseTypeChecker {

    @Override
    protected QualifierHierarchy createQualifierHierarchy() {
        return new JCIPQualifierHierarchy((GraphQualifierHierarchy)super.createQualifierHierarchy());
    }

    private final class JCIPQualifierHierarchy extends GraphQualifierHierarchy {
        public JCIPQualifierHierarchy(GraphQualifierHierarchy hierarchy) {
            super(hierarchy);
        }

        private List<String> guardedByValue(AnnotationMirror anno) {
            AnnotationValue value = anno.getElementValues().values().iterator().next();
            @SuppressWarnings("unchecked")
            List<String> valueList = (List<String>)value.getValue();
            return valueList;
        }

        public boolean isSubtype(AnnotationMirror rhs, AnnotationMirror lhs) {
            if (!AnnotationUtils.areSameIgnoringValues(rhs, lhs))
                return false;
            assert AnnotationUtils.annotationName(rhs).contentEquals(GuardedBy.class.getCanonicalName());

            List<String> rhsLocks = guardedByValue(rhs);
            List<String> lhsLocks = guardedByValue(lhs);

            return lhsLocks.containsAll(rhsLocks);
        }
    }
}
