package checkers.util;

import javax.annotation.processing.ProcessingEnvironment;
import javax.lang.model.element.AnnotationMirror;
import javax.lang.model.element.AnnotationValue;
import javax.lang.model.element.Element;
import javax.lang.model.element.ElementKind;
import javax.lang.model.element.ExecutableElement;
import javax.lang.model.element.Name;
import javax.lang.model.element.TypeElement;
import javax.lang.model.type.DeclaredType;
import javax.lang.model.type.ExecutableType;
import javax.lang.model.type.MirroredTypeException;
import javax.lang.model.type.TypeMirror;
import javax.lang.model.util.ElementFilter;
import javax.lang.model.util.Elements;

import com.sun.source.tree.*;

import checkers.quals.Dependent;
import checkers.types.AnnotatedTypeFactory;
import checkers.types.AnnotatedTypeMirror;
import checkers.types.AnnotatedTypeMirror.AnnotatedExecutableType;

public class DependentTypes {
    private final ProcessingEnvironment env;
    private final Elements elements;
    private final AnnotatedTypeFactory factory;
    private final AnnotationUtils annoUtils;
    private final ExecutableElement resultElement;
    private final ExecutableElement whenElement;

    public DependentTypes(ProcessingEnvironment env, CompilationUnitTree root) {
        this.env = env;
        this.elements = env.getElementUtils();
        this.factory = new AnnotatedTypeFactory(env, null, root);
        this.annoUtils = AnnotationUtils.getInstance(env);

        ExecutableElement result = null, when = null;
        TypeElement depElt = elements.getTypeElement(Dependent.class.getCanonicalName());
        for (ExecutableElement member : ElementFilter.methodsIn(depElt.getEnclosedElements())) {
            if ("when".equals(member.getSimpleName().toString()))
                when = member;
            else if ("result".equals(member.getSimpleName().toString()))
                result = member;
        }
        this.resultElement = result;
        this.whenElement = when;
    }

    AnnotationMirror getResult(Dependent anno) {
        try {
            anno.result();
        } catch (MirroredTypeException exp) {
            Name valName = TypesUtils.getQualifiedName((DeclaredType)exp.getTypeMirror());
            return annoUtils.fromName(valName);
        }
        assert false : "shouldn't be here";
        return null;
    }

    AnnotationMirror getWhen(Dependent anno) {
        try {
            anno.when();
        } catch (MirroredTypeException exp) {
            Name valName = TypesUtils.getQualifiedName((DeclaredType)exp.getTypeMirror());
            return annoUtils.fromName(valName);
        }
        assert false : "shouldn't be here";
        return null;
    }

    public void doSubsitution(Element symbol, AnnotatedTypeMirror type, AnnotatedTypeMirror receiver) {
        Dependent dependentInfo = symbol.getAnnotation(Dependent.class);
        if (dependentInfo == null)
            return;

        if (receiver == null)
            return;

        AnnotationMirror ifpresent = getWhen(dependentInfo);
        if (receiver.hasAnnotation(ifpresent)) {
            type.clearAnnotations();
            AnnotationMirror then = getResult(dependentInfo);
            type.addAnnotation(then);
        }
    }

    public void handle(Tree tree, AnnotatedTypeMirror type) {
        if (!(tree instanceof ExpressionTree))
            return;
        ExpressionTree expr = (ExpressionTree)tree;
        Element symbol = null;
        if (expr instanceof IdentifierTree)
            symbol = TreeUtils.elementFromUse((IdentifierTree)expr);
        else if (expr instanceof MemberSelectTree)
            symbol = TreeUtils.elementFromUse((MemberSelectTree)expr);

        if (symbol == null
                || (!symbol.getKind().isField()
                    && symbol.getKind() != ElementKind.LOCAL_VARIABLE))
            return;

        // FIXME: handle this case
        AnnotatedTypeMirror receiver = factory.getReceiver(expr);
        if (receiver != null)
            doSubsitution(symbol, type, receiver);
    }

    public void handleConstructor(NewClassTree tree, AnnotatedExecutableType type) {
        if (!(tree.getIdentifier() instanceof AnnotatedTypeTree))
            return;
        AnnotatedTypeMirror dt = factory.getAnnotatedType(tree);

        ExecutableElement constructorElt = InternalUtils.constructor(tree);
        for (int i = 0; i < constructorElt.getParameters().size(); ++i) {
            Element parameter = constructorElt.getParameters().get(i);
            AnnotatedTypeMirror paramType = type.getParameterTypes().get(i);
            doSubsitution(parameter, paramType, dt);
        }
    }
}
